@@ -1,5 +1,5 @@
 /* mongo-sync.h - libmongo-client synchronous wrapper API
- * Copyright 2011, 2012 Gergely Nagy <algernon@balabit.hu>
+ * Copyright 2011, 2012, 2013, 2014 Gergely Nagy <algernon@balabit.hu>
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -50,6 +50,57 @@ G_BEGIN_DECLS
 /** Opaque synchronous connection object. */
 typedef struct _mongo_sync_connection mongo_sync_connection;
 
+/** synchronous connection recovery cache object */
+typedef struct _mongo_sync_conn_recovery_cache mongo_sync_conn_recovery_cache;
+
+/** Create a new connection recovery cache object.
+ *
+ * @return the newly created recovery cache object
+ */
+mongo_sync_conn_recovery_cache *mongo_sync_conn_recovery_cache_new (void);
+
+/** Free a connection recovery cache object.
+ *
+ * @param cache is the recovery cache object
+ */
+void mongo_sync_conn_recovery_cache_free (mongo_sync_conn_recovery_cache *cache);
+
+/** Discards a connection recovery cache object.
+ *
+ * @param cache is the recovery cache object
+ */
+void mongo_sync_conn_recovery_cache_discard (mongo_sync_conn_recovery_cache *cache);
+
+/** Add a seed to a connection recovery cache object.
+ *
+ * The seed list will be used for reconnects, prioritized before the
+ * automatically discovered host list.
+ *
+ * @param cache is the connection recovery cache to add a seed to.
+ * @param host is the seed host to add.
+ * @param port is the seed's port.
+ *
+ * @returns TRUE on success, FALSE otherwise.
+ */
+gboolean mongo_sync_conn_recovery_cache_seed_add (mongo_sync_conn_recovery_cache *cache,
+                                                  const gchar *host, gint port);
+
+/** Synchronously connect to a MongoDB server using an external
+ *  connection recovery cache object.
+ *
+ * Sets up a synchronous connection to a MongoDB server.
+ *
+ * @param cache is the externally managed connection recovery cache object.
+ * @param slaveok signals whether queries made against a slave are
+ * acceptable.
+ *
+ * @returns A newly allocated mongo_sync_connection object, or NULL on
+ * error. It is the responsibility of the caller to close and free the
+ * connection when appropriate.
+ */
+mongo_sync_connection *mongo_sync_connect_recovery_cache (mongo_sync_conn_recovery_cache *cache,
+                                                          gboolean slaveok);
+
 /** Synchronously connect to a MongoDB server.
  *
  * Sets up a synchronous connection to a MongoDB server.
@@ -65,8 +116,8 @@ typedef struct _mongo_sync_connection mongo_sync_connection;
  * connection when appropriate.
  */
 mongo_sync_connection *mongo_sync_connect (const gchar *address,
-					   gint port,
-					   gboolean slaveok);
+                                           gint port,
+                                           gboolean slaveok);
 
 /** Add a seed to an existing MongoDB connection.
  *
@@ -80,7 +131,7 @@ mongo_sync_connection *mongo_sync_connect (const gchar *address,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_conn_seed_add (mongo_sync_connection *conn,
-				   const gchar *host, gint port);
+                                   const gchar *host, gint port);
 
 /** Attempt to connect to another member of a replica set.
  *
@@ -97,7 +148,7 @@ gboolean mongo_sync_conn_seed_add (mongo_sync_connection *conn,
  * @note The original connection object will be updated too!
  */
 mongo_sync_connection *mongo_sync_reconnect (mongo_sync_connection *conn,
-					     gboolean force_master);
+                                             gboolean force_master);
 
 /** Close and free a synchronous MongoDB connection.
  *
@@ -123,7 +174,7 @@ gboolean mongo_sync_conn_get_slaveok (const mongo_sync_connection *conn);
  * @returns TRUE on sucess, FALSE otherwise.
  */
 gboolean mongo_sync_conn_set_slaveok (mongo_sync_connection *conn,
-				      gboolean slaveok);
+                                      gboolean slaveok);
 
 /** Retrieve the state of the safe mode flag from a sync connection.
  *
@@ -148,7 +199,7 @@ gboolean mongo_sync_conn_get_safe_mode (const mongo_sync_connection *conn);
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_conn_set_safe_mode (mongo_sync_connection *conn,
-					gboolean safe_mode);
+                                        gboolean safe_mode);
 
 /** Get the state of the auto-reconnect flag from a sync connection.
  *
@@ -174,7 +225,7 @@ gboolean mongo_sync_conn_get_auto_reconnect (const mongo_sync_connection *conn);
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_conn_set_auto_reconnect (mongo_sync_connection *conn,
-					     gboolean auto_reconnect);
+                                             gboolean auto_reconnect);
 
 /** Get the maximum size of a bulk insert package.
  *
@@ -197,7 +248,7 @@ gint32 mongo_sync_conn_get_max_insert_size (mongo_sync_connection *conn);
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_conn_set_max_insert_size (mongo_sync_connection *conn,
-					      gint32 max_size);
+                                              gint32 max_size);
 
 /** Send an update command to MongoDB.
  *
@@ -214,9 +265,9 @@ gboolean mongo_sync_conn_set_max_insert_size (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_update (mongo_sync_connection *conn,
-				const gchar *ns,
-				gint32 flags, const bson *selector,
-				const bson *update);
+                                const gchar *ns,
+                                gint32 flags, const bson *selector,
+                                const bson *update);
 
 /** Send an insert command to MongoDB.
  *
@@ -230,7 +281,7 @@ gboolean mongo_sync_cmd_update (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_insert (mongo_sync_connection *conn,
-				const gchar *ns, ...) G_GNUC_NULL_TERMINATED;
+                                const gchar *ns, ...) G_GNUC_NULL_TERMINATED;
 
 
 /** Send an insert command to MongoDB.
@@ -246,8 +297,8 @@ gboolean mongo_sync_cmd_insert (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_insert_n (mongo_sync_connection *conn,
-				  const gchar *ns, gint32 n,
-				  const bson **docs);
+                                  const gchar *ns, gint32 n,
+                                  const bson **docs);
 
 /** Send a query command to MongoDB.
  *
@@ -266,9 +317,9 @@ gboolean mongo_sync_cmd_insert_n (mongo_sync_connection *conn,
  * anymore.
  */
 mongo_packet *mongo_sync_cmd_query (mongo_sync_connection *conn,
-				    const gchar *ns, gint32 flags,
-				    gint32 skip, gint32 ret, const bson *query,
-				    const bson *sel);
+                                    const gchar *ns, gint32 flags,
+                                    gint32 skip, gint32 ret, const bson *query,
+                                    const bson *sel);
 
 /** Send a get more command to MongoDB.
  *
@@ -283,8 +334,8 @@ mongo_packet *mongo_sync_cmd_query (mongo_sync_connection *conn,
  * used anymore.
  */
 mongo_packet *mongo_sync_cmd_get_more (mongo_sync_connection *conn,
-				       const gchar *ns,
-				       gint32 ret, gint64 cursor_id);
+                                       const gchar *ns,
+                                       gint32 ret, gint64 cursor_id);
 
 /** Send a delete command to MongoDB.
  *
@@ -297,7 +348,7 @@ mongo_packet *mongo_sync_cmd_get_more (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_delete (mongo_sync_connection *conn, const gchar *ns,
-				gint32 flags, const bson *sel);
+                                gint32 flags, const bson *sel);
 
 /** Send a kill_cursors command to MongoDB.
  *
@@ -310,7 +361,7 @@ gboolean mongo_sync_cmd_delete (mongo_sync_connection *conn, const gchar *ns,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_kill_cursors (mongo_sync_connection *conn,
-				      gint32 n, ...);
+                                      gint32 n, ...);
 
 /** Send a custom command to MongoDB.
  *
@@ -326,8 +377,8 @@ gboolean mongo_sync_cmd_kill_cursors (mongo_sync_connection *conn,
  * used anymore.
  */
 mongo_packet *mongo_sync_cmd_custom (mongo_sync_connection *conn,
-				     const gchar *db,
-				     const bson *command);
+                                     const gchar *db,
+                                     const bson *command);
 
 /** Send a count() command to MongoDB.
  *
@@ -343,8 +394,8 @@ mongo_packet *mongo_sync_cmd_custom (mongo_sync_connection *conn,
  * @returns The number of matching documents, or -1 on error.
  */
 gdouble mongo_sync_cmd_count (mongo_sync_connection *conn,
-			      const gchar *db, const gchar *coll,
-			      const bson *query);
+                              const gchar *db, const gchar *coll,
+                              const bson *query);
 
 /** Flags that can be set during collection creation. */
 enum
@@ -383,8 +434,8 @@ enum
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_create (mongo_sync_connection *conn,
-				const gchar *db, const gchar *coll,
-				gint flags, ...);
+                                const gchar *db, const gchar *coll,
+                                gint flags, ...);
 
 /** Check whether a collection exists in MongoDB.
  *
@@ -397,7 +448,7 @@ gboolean mongo_sync_cmd_create (mongo_sync_connection *conn,
  * the caller to free the BSON object once it is no longer needed.
  */
 bson *mongo_sync_cmd_exists (mongo_sync_connection *conn,
-			     const gchar *db, const gchar *coll);
+                             const gchar *db, const gchar *coll);
 
 /** Send a drop() command to MongoDB.
  *
@@ -410,7 +461,7 @@ bson *mongo_sync_cmd_exists (mongo_sync_connection *conn,
  * @returns TRUE if the collection was dropped, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_drop (mongo_sync_connection *conn,
-			      const gchar *db, const gchar *coll);
+                              const gchar *db, const gchar *coll);
 
 /** Get the last error from MongoDB.
  *
@@ -426,7 +477,23 @@ gboolean mongo_sync_cmd_drop (mongo_sync_connection *conn,
  * is returning TRUE.
  */
 gboolean mongo_sync_cmd_get_last_error (mongo_sync_connection *conn,
-					const gchar *db, gchar **error);
+                                        const gchar *db, gchar **error);
+
+/** Get the last error from MongoDB.
+ *
+ * Retrieves the last error from MongoDB.
+ *
+ * @param conn is the connection to work with.
+ * @param db is the name of the database.
+ * @param error is a pointer to a BSON variable that will hold the
+ * error message.
+ *
+ * @returns TRUE if the error was succesfully retrieved, FALSE
+ * otherwise. The output variable @a error is only set if the function
+ * is returning TRUE.
+ */
+gboolean mongo_sync_cmd_get_last_error_full (mongo_sync_connection *conn,
+                                             const gchar *db, bson **error);
 
 /** Reset the last error variable in MongoDB.
  *
@@ -436,7 +503,7 @@ gboolean mongo_sync_cmd_get_last_error (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_reset_error (mongo_sync_connection *conn,
-				     const gchar *db);
+                                     const gchar *db);
 
 /** Check whether the current node is the master.
  *
@@ -465,9 +532,25 @@ gboolean mongo_sync_cmd_ping (mongo_sync_connection *conn);
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_user_add (mongo_sync_connection *conn,
-				  const gchar *db,
-				  const gchar *user,
-				  const gchar *pw);
+                                  const gchar *db,
+                                  const gchar *user,
+                                  const gchar *pw);
+
+/** Add a user to MongoDB, with roles.
+ *
+ * @param conn is the connection to work with.
+ * @param db is the database to add the user to.
+ * @param user is the user to add.
+ * @param pw is the password.
+ * @param roles is a BSON array containing the roles for the user.
+ *
+ * @returns TRUE on success, FALSE otherwise.
+ */
+gboolean mongo_sync_cmd_user_add_with_roles (mongo_sync_connection *conn,
+                                             const gchar *db,
+                                             const gchar *user,
+                                             const gchar *pw,
+                                             const bson *roles);
 
 /** Remove a user from MongoDB.
  *
@@ -478,8 +561,8 @@ gboolean mongo_sync_cmd_user_add (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_user_remove (mongo_sync_connection *conn,
-				     const gchar *db,
-				     const gchar *user);
+                                     const gchar *db,
+                                     const gchar *user);
 
 /** Authenticate a user with MongoDB.
  *
@@ -491,18 +574,18 @@ gboolean mongo_sync_cmd_user_remove (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_authenticate (mongo_sync_connection *conn,
-				      const gchar *db,
-				      const gchar *user,
-				      const gchar *pw);
+                                      const gchar *db,
+                                      const gchar *user,
+                                      const gchar *pw);
 
 /** Flags that can be set at index creation. */
 enum
   {
     MONGO_INDEX_UNIQUE = 0x01, /**< Create a unique index. */
     MONGO_INDEX_DROP_DUPS = 0x02, /**< Drop duplicate entries when
-				     creating the indexes. */
+                                     creating the indexes. */
     MONGO_INDEX_BACKGROUND = 0x04, /**< Create indexes in the
-				      background. */
+                                      background. */
     MONGO_INDEX_SPARSE = 0x08 /**< Create sparse indexes. */
   };
 
@@ -516,9 +599,9 @@ enum
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_index_create (mongo_sync_connection *conn,
-				      const gchar *ns,
-				      const bson *key,
-				      gint options);
+                                      const gchar *ns,
+                                      const bson *key,
+                                      gint options);
 
 /** Drop an index.
  *
@@ -529,8 +612,8 @@ gboolean mongo_sync_cmd_index_create (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_index_drop (mongo_sync_connection *conn,
-				    const gchar *ns,
-				    const bson *key);
+                                    const gchar *ns,
+                                    const bson *key);
 
 /** Drop all indexes from a namespace.
  *
@@ -540,7 +623,15 @@ gboolean mongo_sync_cmd_index_drop (mongo_sync_connection *conn,
  * @returns TRUE on success, FALSE otherwise.
  */
 gboolean mongo_sync_cmd_index_drop_all (mongo_sync_connection *conn,
-					const gchar *ns);
+                                        const gchar *ns);
+
+/** Get the last error message on a connection
+ *
+ * @param conn is the connection
+ *
+ * @returns pointer to the error message, if exists, NULL otherwise
+ */
+const gchar *mongo_sync_conn_get_last_error (mongo_sync_connection *conn);
 
 /** @} */
 
