@@ -28,6 +28,8 @@
 #include <unistd.h>
 #include <errno.h>
 
+#include "mongo-client.h"
+
 static guint32 machine_id = 0;
 static gint16 pid = 0;
 
@@ -136,7 +138,10 @@ mongo_util_parse_addr (const gchar *addr, gchar **host, gint *port)
 
       port_s += 2;
       if (port_s - addr >= (glong)strlen (addr))
-        return TRUE;
+        {
+          *port = -1;
+          return TRUE;
+        }
     }
   else
     {
@@ -147,6 +152,7 @@ mongo_util_parse_addr (const gchar *addr, gchar **host, gint *port)
       if (!port_s)
         {
           *host = g_strdup (addr);
+          *port = -1;
           return TRUE;
         }
       if (port_s == addr)
@@ -169,7 +175,7 @@ mongo_util_parse_addr (const gchar *addr, gchar **host, gint *port)
       errno = ERANGE;
       return FALSE;
     }
-  if (p < 0 || p > INT_MAX)
+  if ((p != MONGO_CONN_LOCAL) && (p < 0 || p > INT_MAX))
     {
       g_free (*host);
       *host = NULL;
